@@ -5,9 +5,7 @@
  */
 
 #include "fd.h"
-#include "term.h"
 #include "func.h"
-#include "kctype.h"
 #include "kanji.h"
 
 #ifndef	_NOARCHIVE
@@ -16,15 +14,15 @@
 #include "system.h"
 #endif
 
-#ifndef	_NODOSDRIVE
-extern int shutdrv __P_((int));
-#endif
 #if	MSDOS
 extern char *unixgetcurdir __P_((char *, int));
 #endif
 
 extern int mark;
 extern int stackdepth;
+#ifndef	_NOTRADLAYOUT
+extern int tradlayout;
+#endif
 extern int sizeinfo;
 extern off_t marksize;
 extern off_t totalsize;
@@ -68,7 +66,7 @@ static char *ign_lha[] = {
 #else	/* !MSDOS */
 	" PERMSSN * UID*GID *",
 	"----------*",
-	" Total * file* ???.?%*",
+	" Total * file* ???.*%*",
 #endif	/* !MSDOS */
 	NULL
 };
@@ -147,7 +145,7 @@ static char *form_tar[] = {
 		{0, 0, 0, 0, 0, 0, 0, 0, 0}, \
 		{255, 255, 255}, 1
 #define	LINESEP	'\t'
-#define	isarchbr(l)	(((l) -> topskip) < 255)
+#define	isarchbr(l)	(((l) -> topskip) < (u_char)255)
 #endif	/* FD < 2 */
 
 #define	iswhitespace(c)	((c) == ' ' || (c) == '\t')
@@ -580,14 +578,12 @@ static char *NEAR readfname(s, len)
 char *s;
 int len;
 {
-	char buf[MAXPATHLEN];
+	char *cp, *tmp;
 
-	s = strndup2(s, len);
-	if (kanjiconv2(buf, s, MAXPATHLEN - 1, fnamekcode, DEFCODE) != s) {
-		free(s);
-		s = strdup2(buf);
-	}
-	return(s);
+	cp = strndup2(s, len);
+	tmp = newkanjiconv(cp, fnamekcode, DEFCODE, L_FNAME);
+	if (tmp != cp) free(cp);
+	return(tmp);
 }
 # endif	/* !_NOKANJIFCONV */
 
@@ -600,8 +596,10 @@ int skip;
 #ifndef	NOUID
 	uidtable *up;
 	gidtable *gp;
+	uid_t uid;
+	gid_t gid;
 #endif
-	long n;
+	off_t n;
 	int i, ch, l, len, hit, err, err2, score;
 	char *cp, *s, *buf, *rawbuf;
 
@@ -615,7 +613,7 @@ int skip;
 	tmp -> st_gid = (gid_t)-1;
 	tmp -> linkname = NULL;
 #endif
-	tmp -> st_size = 0L;
+	tmp -> st_size = (off_t)0;
 	tmp -> flags = 0;
 	tmp -> tmpflags = F_STAT;
 	tm.tm_year = tm.tm_mon = tm.tm_mday = -1;
@@ -647,10 +645,10 @@ int skip;
 			for (i = 0; line[i]; i++) if (line[i] == '\n') break;
 			len = i;
 		}
-		else if (!(cp = evalnumeric(form, &n, 1))) len = -1;
+		else if (!(cp = sscanf2(form, "%+d", &len))) len = -1;
 		else {
 			form = cp;
-			for (i = 0; i < n; i++) if (!line[i]) break;
+			for (i = 0; i < len; i++) if (!line[i]) break;
 			len = i;
 			ch = -1;
 		}
@@ -702,8 +700,9 @@ int skip;
 				break;
 			case 'u':
 #ifndef	NOUID
-				if ((cp = evalnumeric(buf, &n, -1)) && !*cp)
-					tmp -> st_uid = n;
+				cp = sscanf2(buf, "%-*d%$",
+					sizeof(uid_t), &uid);
+				if (cp) tmp -> st_uid = uid;
 				else if ((up = finduid(0, buf)))
 					tmp -> st_uid = up -> uid;
 #endif
@@ -711,15 +710,16 @@ int skip;
 				break;
 			case 'g':
 #ifndef	NOUID
-				if ((cp = evalnumeric(buf, &n, -1)) && !*cp)
-					tmp -> st_gid = n;
+				cp = sscanf2(buf, "%-*d%$",
+					sizeof(gid_t), &gid);
+				if (cp) tmp -> st_gid = gid;
 				else if ((gp = findgid(0, buf)))
 					tmp -> st_gid = gp -> gid;
 #endif
 				hit++;
 				break;
 			case 's':
-				if (!(cp = evalnumeric(buf, &n, 0))) break;
+				if (!(cp = sscanf2(buf, "%qd", &n))) break;
 				if (*cp) err2++;
 				tmp -> st_size = n;
 				hit++;
@@ -764,28 +764,21 @@ int skip;
 				hit++;
 				break;
 			case 't':
-				if (!(cp = evalnumeric(buf, &n, 0))
-				|| n > 23) {
-					tm.tm_hour = tm.tm_min = tm.tm_sec = 0;
+				if (!(cp = sscanf2(buf, "%d", &i)) || i > 23)
 					break;
-				}
-				tm.tm_hour = n;
+				tm.tm_hour = i;
 				hit++;
-				if (*cp != ':'
-				|| !(cp = evalnumeric(++cp, &n, 0))
-				|| n > 59) {
+				if (!(cp = sscanf2(cp, ":%d", &i)) || i > 59) {
 					tm.tm_min = tm.tm_sec = 0;
 					err2++;
 					break;
 				}
-				tm.tm_min = n;
-				if (*cp != ':'
-				|| !(cp = evalnumeric(++cp, &n, 0))
-				|| n > 59) {
+				tm.tm_min = i;
+				if (!sscanf2(cp, ":%d%$", &i) || i > 59) {
 					tm.tm_sec = 0;
 					break;
 				}
-				tm.tm_sec = n;
+				tm.tm_sec = i;
 				break;
 			case 'f':
 				for (i = 0; i < len; i++) {
@@ -918,12 +911,12 @@ int field, *eolp;
 			j++;
 			sp = (j < MAXLAUNCHSEP) ? (int)sep[j] : 255;
 			for (; sp == 255 || i < sp; i++)
-				if (!strchr(" \t", line[i])) break;
+				if (!iswhitespace(line[i])) break;
 			if (f < 0) f = 1;
 			else f++;
 			s = 0;
 		}
-		else if (strchr(" \t", line[i])) s = 1;
+		else if (iswhitespace(line[i])) s = 1;
 		else if (s) {
 			f++;
 			s = 0;
@@ -935,7 +928,7 @@ int field, *eolp;
 			if (eolp) {
 				for (j = i; line[j]; j++) {
 					if ((sp < 255 && j >= sp)
-					|| strchr(" \t", line[j])) break;
+					|| iswhitespace(line[j])) break;
 				}
 				*eolp = j;
 			}
@@ -992,8 +985,10 @@ int max;
 #ifndef	NOUID
 	uidtable *up;
 	gidtable *gp;
+	uid_t uid;
+	gid_t gid;
 #endif
-	long n;
+	off_t n;
 	int i, skip;
 	char *cp, *buf;
 
@@ -1037,14 +1032,14 @@ int max;
 
 #ifndef	NOUID
 	getfield(buf, line, skip, list, F_UID);
-	if ((cp = evalnumeric(buf, &n, -1)) && !*cp) tmp -> st_uid = n;
+	if (sscanf2(buf, "%-*d%$", sizeof(uid_t), &uid)) tmp -> st_uid = uid;
 	else tmp -> st_uid = ((up = finduid(0, buf))) ? up -> uid : (uid_t)-1;
 	getfield(buf, line, skip, list, F_GID);
-	if ((cp = evalnumeric(buf, &n, -1)) && !*cp) tmp -> st_gid = n;
+	if (sscanf2(buf, "%-*d%$", sizeof(gid_t), &gid)) tmp -> st_gid = gid;
 	else tmp -> st_gid = ((gp = findgid(0, buf))) ? gp -> gid : (gid_t)-1;
 #endif
 	getfield(buf, line, skip, list, F_SIZE);
-	tmp -> st_size = ((cp = evalnumeric(buf, &n, 0)) && !*cp) ? n : 0L;
+	tmp -> st_size = (sscanf2(buf, "%qd%$", &n)) ? n : (off_t)0;
 
 	getfield(buf, line, skip, list, F_MON);
 	if (!strncmp(buf, "Jan", 3)) tm.tm_mon = 0;
@@ -1065,7 +1060,7 @@ int max;
 	tm.tm_mday = ((i = atoi2(buf)) >= 1 && i < 31) ? i : 1;
 	getfield(buf, line, skip, list, F_YEAR);
 	if (!*buf) tm.tm_year = 1970;
-	else if ((cp = evalnumeric(buf, &n, 0)) && !*cp) tm.tm_year = n;
+	else if ((i = atoi2(buf)) >= 0) tm.tm_year = i;
 	else if (list -> field[F_YEAR] == list -> field[F_TIME]
 	&& strchr(buf, ':')) {
 		tm.tm_year = today[0];
@@ -1079,17 +1074,16 @@ int max;
 	tm.tm_year -= 1900;
 
 	getfield(buf, line, skip, list, F_TIME);
-	if (!(cp = evalnumeric(buf, &n, 0)) || n > 23)
+	if (!(cp = sscanf2(buf, "%d", &i)) || i > 23)
 		tm.tm_hour = tm.tm_min = tm.tm_sec = 0;
 	else {
-		tm.tm_hour = n;
-		if (*cp != ':' || !(cp = evalnumeric(++cp, &n, 0)) || n > 59)
+		tm.tm_hour = i;
+		if (!(cp = sscanf2(cp, ":%d", &i)) || i > 59)
 			tm.tm_min = tm.tm_sec = 0;
 		else {
-			tm.tm_min = n;
-			if (*cp != ':' || (n = atoi2(++cp)) < 0 || n > 59)
-				tm.tm_sec = 0;
-			else tm.tm_sec = n;
+			tm.tm_min = i;
+			if (!sscanf2(cp, ":%d%$", &i) || i > 59) tm.tm_sec = 0;
+			else tm.tm_sec = i;
 		}
 	}
 
@@ -1139,27 +1133,58 @@ char *file, *dir;
 		strcpy(strcpy2(strcatdelim2(arch, fullpath, file), ":"), dir);
 	}
 
+#ifndef	_NOTRADLAYOUT
+	if (istradlayout()) {
+		locate(0, TL_PATH);
+		putterm(l_clear);
+
+		locate(TC_PATH, TL_PATH);
+		putterm(t_standout);
+#ifndef	_NOBROWSE
+		if (browselist) {
+			cputs2(TS_BROWSE);
+			len = TD_BROWSE;
+		}
+		else
+#endif
+		{
+			cputs2(TS_ARCH);
+			len = TD_ARCH;
+		}
+		putterm(end_standout);
+		cprintf2("%-*.*k", len, len, arch);
+		free(arch);
+
+		locate(TC_MARK, TL_PATH);
+		cprintf2("%<*d", TD_MARK, mark);
+		putterm(t_standout);
+		cputs2(TS_MARK);
+		putterm(end_standout);
+		cprintf2("%<'*qd", TD_SIZE, marksize);
+
+		tflush();
+		return;
+	}
+#endif	/* !_NOTRADLAYOUT */
+
 	locate(0, L_PATH);
 	putterm(l_clear);
-	len = 0;
-	if (!isleftshift()) {
-		putch2(' ');
-		len++;
-	}
+
+	locate(C_PATH, L_PATH);
 	putterm(t_standout);
 #ifndef	_NOBROWSE
 	if (browselist) {
-		cputs2("Browse:");
-		len += 7;
+		cputs2(S_BROWSE);
+		len = D_BROWSE;
 	}
 	else
 #endif
 	{
-		cputs2("Arch:");
-		len += 5;
+		cputs2(S_ARCH);
+		len = D_ARCH;
 	}
 	putterm(end_standout);
-	kanjiputs2(arch, n_column - len, 0);
+	cprintf2("%-*.*k", len, len, arch);
 	free(arch);
 
 	tflush();
@@ -1323,10 +1348,10 @@ char *s, **argv;
 }
 #endif	/* FD >= 2 */
 
-static int NEAR parsearchive(fp, list, flist, linenop)
+static int NEAR parsearchive(fp, list, namep, linenop)
 FILE *fp;
 launchtable *list;
-namelist *flist;
+namelist *namep;
 int *linenop;
 {
 #if	FD >= 2
@@ -1362,7 +1387,7 @@ int *linenop;
 	}
 
 #if	FD >= 2
-	if (!formlist) return(-3);
+	if (!formlist || !*formlist) return(-3);
 #endif
 	if (!(nline = countvar(lvar))) {
 		lvar = (char **)realloc2(lvar, (nline + 2) * sizeof(char *));
@@ -1395,9 +1420,9 @@ int *linenop;
 		if (list -> field[i] < 255 && (int)(list -> field[i]) > max)
 			max = (int)(list -> field[i]);
 #endif	/* FD < 2 */
-	flist -> name = NULL;
+	namep -> name = NULL;
 #ifndef	NOSYMLINK
-	flist -> linkname = NULL;
+	namep -> linkname = NULL;
 #endif
 
 	needline = 0;
@@ -1438,11 +1463,11 @@ int *linenop;
 						formlist[i] = formlist[i + 1];
 					continue;
 				}
-				if (flist -> name) free(flist -> name);
-				flist -> name = NULL;
+				if (namep -> name) free(namep -> name);
+				namep -> name = NULL;
 # ifndef	NOSYMLINK
-				if (flist -> linkname) free(flist -> linkname);
-				flist -> linkname = NULL;
+				if (namep -> linkname) free(namep -> linkname);
+				namep -> linkname = NULL;
 # endif
 				if (form0) free(form0);
 				free(scorelist);
@@ -1497,11 +1522,11 @@ int *linenop;
 
 		if (score >= MAXSCORE);
 		else if (!i) {
-			if (flist -> name) free(flist -> name);
+			if (namep -> name) free(namep -> name);
 # ifndef	NOSYMLINK
-			if (flist -> linkname) free(flist -> linkname);
+			if (namep -> linkname) free(namep -> linkname);
 # endif
-			memcpy((char *)flist, (char *)&tmp, sizeof(tmp));
+			memcpy((char *)namep, (char *)&tmp, sizeof(tmp));
 		}
 		else {
 			if (tmp.name) free(tmp.name);
@@ -1512,7 +1537,7 @@ int *linenop;
 
 		if (!score) break;
 #else	/* FD < 2 */
-		score = readfileent(flist, cp, list, max);
+		score = readfileent(namep, cp, list, max);
 		free(cp);
 		break;
 /*NOTREACHED*/
@@ -1521,11 +1546,11 @@ int *linenop;
 
 #if	FD >= 2
 	if (score < 0) {
-		if (flist -> name) free(flist -> name);
-		flist -> name = NULL;
+		if (namep -> name) free(namep -> name);
+		namep -> name = NULL;
 # ifndef	NOSYMLINK
-		if (flist -> linkname) free(flist -> linkname);
-		flist -> linkname = NULL;
+		if (namep -> linkname) free(namep -> linkname);
+		namep -> linkname = NULL;
 # endif
 	}
 #endif	/* FD >= 2 */
@@ -1548,16 +1573,15 @@ int *linenop;
 
 static VOID NEAR unpackerror(VOID_A)
 {
-	if (isttyiomode) {
-		locate(0, n_line - 1);
-		cputs2("\r\n");
-		tflush();
-		hideclock = 1;
-		warning(0, UNPNG_K);
+	if (!isttyiomode) {
+		kanjifputs(UNPNG_K, stderr);
+		fputnl(stderr);
 	}
 	else {
-		fputs(UNPNG_K, stderr);
-		fputc('\n', stderr);
+		locate(0, n_line - 1);
+		cputnl();
+		hideclock = 1;
+		warning(0, UNPNG_K);
 	}
 }
 
