@@ -11,6 +11,9 @@
 #include <string.h>
 #include <errno.h>
 #include "machine.h"
+#include "printf.h"
+#include "kctype.h"
+#include "pathname.h"
 
 # ifndef	NOUNISTDH
 # include <unistd.h>
@@ -20,12 +23,12 @@
 # include <stdlib.h>
 # endif
 #endif	/* !FD */
-#include <ctype.h>
 
 #if	MSDOS
 #include <process.h>
 #include "unixemu.h"
 # ifndef	FD
+# include <fcntl.h>
 # include <dos.h>
 # include <io.h>
 # define	DS_IRDONLY	001
@@ -38,13 +41,19 @@
 			| DS_IFDIR | DS_IARCHIVE)
 # endif	/* !FD */
 #define	EXTWIDTH	4
-# if	defined (DJGPP) && DJGPP < 2
+# if	defined (__TURBOC__) || (defined (DJGPP) && DJGPP < 2)
 # include <dir.h>
+# endif
+# if	defined (DJGPP) && DJGPP < 2
 # define	find_t	ffblk
 # define	_dos_findfirst(p, a, f)	findfirst(p, f, a)
 # define	_dos_findnext		findnext
 # define	_ENOENT_		ENMFILE
 # else
+# define	ff_attrib		attrib
+# define	ff_ftime		wr_time
+# define	ff_fdate		wr_date
+# define	ff_fsize		size
 # define	ff_name			name
 # define	_ENOENT_		ENOENT
 # endif
@@ -62,8 +71,6 @@
 # endif
 #endif	/* !MSDOS */
 
-#include "pathname.h"
-
 #ifdef	NOERRNO
 extern int errno;
 #endif
@@ -78,9 +85,9 @@ extern char *realloc2 __P_((VOID_P, ALLOC_T));
 extern char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
 extern char *strdup2 __P_((char *));
 extern char *strndup2 __P_((char *, int));
+extern char *strchr2 __P_((char *, int));
 extern char *strcpy2 __P_((char *, char *));
 extern char *strncpy2 __P_((char *, char *, int));
-extern int snprintf2 __P_((char *, int, CONST char *, ...));
 # ifdef	_USEDOSPATH
 extern int _dospath __P_((char *));
 # endif
@@ -97,39 +104,42 @@ extern VOID demacroarg __P_((char **));
 extern char *progpath;
 #else	/* !FD */
 #define	getenv2		(char *)getenv
-static VOID NEAR allocerror __P_((VOID_A));
+VOID error __P_((char *));
 char *malloc2 __P_((ALLOC_T));
 char *realloc2 __P_((VOID_P, ALLOC_T));
 char *c_realloc __P_((char *, ALLOC_T, ALLOC_T *));
 char *strdup2 __P_((char *));
 char *strndup2 __P_((char *, int));
+char *strchr2 __P_((char *, int));
 char *strcpy2 __P_((char *, char *));
 char *strncpy2 __P_((char *, char *, int));
-char *ascnumeric __P_((char *, off_t, int, int));
 # if	MSDOS
-# define	_dospath(s)	(isalpha(*(s)) && (s)[1] == ':')
+int _dospath __P_((char *));
 DIR *Xopendir __P_((char *));
 int Xclosedir __P_((DIR *));
 struct dirent *Xreaddir __P_((DIR *));
-# else
+static u_int NEAR getdosmode __P_((u_int));
+static time_t NEAR getdostime __P_((u_int, u_int));
+int Xstat __P_((char *, struct stat *));
+# else	/* !MSDOS */
 # define	Xopendir	opendir
 # define	Xclosedir	closedir
 # define	Xreaddir	readdir
-# endif
+# define	Xstat		stat
+# endif	/* !MSDOS */
 # ifdef	DJGPP
 char *Xgetwd __P_((char *));
-# else
+# else	/* !DJGPP */
 #  ifdef	USEGETWD
 #  define	Xgetwd		(char *)getwd
 #  else
 #  define	Xgetwd(p)	(char *)getcwd(p, MAXPATHLEN)
 #  endif
-# endif
-#define	Xstat(f, s)	(stat(f, s) ? -1 : 0)
+# endif	/* !DJGPP */
 #define	Xaccess		access
 #endif	/* !FD */
 
-static char *NEAR getvar __P_((char *, int));
+static char *NEAR getenvvar __P_((char *, int));
 static int NEAR setvar __P_((char *, char *, int));
 static int NEAR ismeta __P_((char *s, int, int, int));
 #ifdef	_NOORIGGLOB
@@ -180,44 +190,12 @@ static char *NEAR insertarg __P_((char *, int, char *, int, int));
 static int NEAR evalvar __P_((char **, int, char **, int));
 static char *NEAR replacebackquote __P_((char *, int *, char *, int));
 
-#ifdef	LSI_C
-#include <jctype.h>
-#define	iskna(c)	iskana(c)
-#define	issjis1(c)	iskanji(c)
-#define	issjis2(c)	iskanji2(c)
-#else	/* !LSI_C */
-#define	KC_SJIS1	0001
-#define	KC_SJIS2	0002
-#define	KC_KANA		0004
-#define	KC_EUCJP	0010
-#define	iskna(c)	(kctypetable[(u_char)(c)] & KC_KANA)
-
-# ifndef	issjis1
-# define	issjis1(c)	(kctypetable[(u_char)(c)] & KC_SJIS1)
-# endif
-# ifndef	issjis2
-# define	issjis2(c)	(kctypetable[(u_char)(c)] & KC_SJIS2)
-# endif
-
-# ifndef	iseuc
-# define	iseuc(c)	(kctypetable[(u_char)(c)] & KC_EUCJP)
-# endif
-#endif	/* !LSI_C */
-
-#define	isekana(s, i)	((u_char)((s)[i]) == 0x8e && iskna((s)[(i) + 1]))
-
-#ifdef	CODEEUC
-#define	iskanji1(s, i)	(iseuc((s)[i]) && iseuc((s)[(i) + 1]))
-#else
-#define	iskanji1(s, i)	(issjis1((s)[i]) && issjis2((s)[(i) + 1]))
-#endif
-
 #define	BUFUNIT		32
-#define	b_size(n, type) ((((n) / BUFUNIT) + 1) * BUFUNIT * sizeof(type))
+#define	b_size(n, type)	((((n) / BUFUNIT) + 1) * BUFUNIT * sizeof(type))
 #define	b_realloc(ptr, n, type) \
 			(((n) % BUFUNIT) ? ((type *)(ptr)) \
 			: (type *)realloc2(ptr, b_size(n, type)))
-#define	getconstvar(s)	(getvar(s, sizeof(s) - 1))
+#define	getconstvar(s)	(getenvvar(s, sizeof(s) - 1))
 
 char **argvar = NULL;
 #ifndef	_NOUSEHASH
@@ -238,114 +216,6 @@ char *(*posixsubstfunc)__P_((char *, int *)) = NULL;
 #ifndef	PATHNOCASE
 int pathignorecase = 0;
 #endif
-#ifndef	LSI_C
-u_char uppercase[256] = {
-	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* 0x00 */
-	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
-	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,	/* 0x10 */
-	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
-	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,	/* 0x20 */
-	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
-	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,	/* 0x30 */
-	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
-	0x40, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,	/* 0x40 */
-	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
-	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,	/* 0x50 */
-	0x58, 0x59, 0x5a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
-	0x60, 0x41, 0x42, 0x43, 0x44, 0x45, 0x46, 0x47,	/* 0x60 */
-	0x48, 0x49, 0x4a, 0x4b, 0x4c, 0x4d, 0x4e, 0x4f,
-	0x50, 0x51, 0x52, 0x53, 0x54, 0x55, 0x56, 0x57,	/* 0x70 */
-	0x58, 0x59, 0x5a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
-	0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,	/* 0x80 */
-	0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
-	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,	/* 0x90 */
-	0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
-	0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,	/* 0xa0 */
-	0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
-	0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7,	/* 0xb0 */
-	0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
-	0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,	/* 0xc0 */
-	0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
-	0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,	/* 0xd0 */
-	0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
-	0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,	/* 0xe0 */
-	0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
-	0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7,	/* 0xf0 */
-	0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
-};
-u_char lowercase[256] = {
-	0x00, 0x01, 0x02, 0x03, 0x04, 0x05, 0x06, 0x07,	/* 0x00 */
-	0x08, 0x09, 0x0a, 0x0b, 0x0c, 0x0d, 0x0e, 0x0f,
-	0x10, 0x11, 0x12, 0x13, 0x14, 0x15, 0x16, 0x17,	/* 0x10 */
-	0x18, 0x19, 0x1a, 0x1b, 0x1c, 0x1d, 0x1e, 0x1f,
-	0x20, 0x21, 0x22, 0x23, 0x24, 0x25, 0x26, 0x27,	/* 0x20 */
-	0x28, 0x29, 0x2a, 0x2b, 0x2c, 0x2d, 0x2e, 0x2f,
-	0x30, 0x31, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37,	/* 0x30 */
-	0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x3e, 0x3f,
-	0x40, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,	/* 0x40 */
-	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
-	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,	/* 0x50 */
-	0x78, 0x79, 0x7a, 0x5b, 0x5c, 0x5d, 0x5e, 0x5f,
-	0x60, 0x61, 0x62, 0x63, 0x64, 0x65, 0x66, 0x67,	/* 0x60 */
-	0x68, 0x69, 0x6a, 0x6b, 0x6c, 0x6d, 0x6e, 0x6f,
-	0x70, 0x71, 0x72, 0x73, 0x74, 0x75, 0x76, 0x77,	/* 0x70 */
-	0x78, 0x79, 0x7a, 0x7b, 0x7c, 0x7d, 0x7e, 0x7f,
-	0x80, 0x81, 0x82, 0x83, 0x84, 0x85, 0x86, 0x87,	/* 0x80 */
-	0x88, 0x89, 0x8a, 0x8b, 0x8c, 0x8d, 0x8e, 0x8f,
-	0x90, 0x91, 0x92, 0x93, 0x94, 0x95, 0x96, 0x97,	/* 0x90 */
-	0x98, 0x99, 0x9a, 0x9b, 0x9c, 0x9d, 0x9e, 0x9f,
-	0xa0, 0xa1, 0xa2, 0xa3, 0xa4, 0xa5, 0xa6, 0xa7,	/* 0xa0 */
-	0xa8, 0xa9, 0xaa, 0xab, 0xac, 0xad, 0xae, 0xaf,
-	0xb0, 0xb1, 0xb2, 0xb3, 0xb4, 0xb5, 0xb6, 0xb7,	/* 0xb0 */
-	0xb8, 0xb9, 0xba, 0xbb, 0xbc, 0xbd, 0xbe, 0xbf,
-	0xc0, 0xc1, 0xc2, 0xc3, 0xc4, 0xc5, 0xc6, 0xc7,	/* 0xc0 */
-	0xc8, 0xc9, 0xca, 0xcb, 0xcc, 0xcd, 0xce, 0xcf,
-	0xd0, 0xd1, 0xd2, 0xd3, 0xd4, 0xd5, 0xd6, 0xd7,	/* 0xd0 */
-	0xd8, 0xd9, 0xda, 0xdb, 0xdc, 0xdd, 0xde, 0xdf,
-	0xe0, 0xe1, 0xe2, 0xe3, 0xe4, 0xe5, 0xe6, 0xe7,	/* 0xe0 */
-	0xe8, 0xe9, 0xea, 0xeb, 0xec, 0xed, 0xee, 0xef,
-	0xf0, 0xf1, 0xf2, 0xf3, 0xf4, 0xf5, 0xf6, 0xf7,	/* 0xf0 */
-	0xf8, 0xf9, 0xfa, 0xfb, 0xfc, 0xfd, 0xfe, 0xff
-};
-# ifdef	FD
-extern u_char kctypetable[256];
-# else	/* !FD */
-u_char kctypetable[256] = {
-	0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,	/* 0x00 */
-	0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,
-	0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,	/* 0x10 */
-	0200, 0200, 0200, 0200, 0200, 0200, 0200, 0200,
-	   0, 0060, 0060, 0060, 0060, 0060, 0060, 0060,	/* 0x20 */
-	0060, 0060, 0060, 0060, 0060, 0060, 0060, 0060,
-	0060, 0060, 0060, 0060, 0060, 0060, 0060, 0060,	/* 0x30 */
-	0060, 0060, 0060, 0060, 0060, 0060, 0060, 0060,
-	0062, 0062, 0062, 0062, 0062, 0062, 0062, 0062,	/* 0x40 */
-	0062, 0062, 0062, 0062, 0062, 0062, 0062, 0062,
-	0062, 0062, 0062, 0062, 0062, 0062, 0062, 0062,	/* 0x50 */
-	0062, 0062, 0062, 0062, 0062, 0062, 0062, 0062,
-	0022, 0022, 0022, 0022, 0022, 0022, 0022, 0022,	/* 0x60 */
-	0022, 0022, 0022, 0022, 0022, 0022, 0022, 0022,
-	0022, 0022, 0022, 0022, 0022, 0022, 0022, 0022,	/* 0x70 */
-	0022, 0022, 0022, 0022, 0022, 0022, 0022, 0200,
-	0002, 0003, 0003, 0003, 0003, 0003, 0003, 0003,	/* 0x80 */
-	0003, 0003, 0003, 0003, 0003, 0003, 0003, 0003,
-	0003, 0003, 0003, 0003, 0003, 0003, 0003, 0003,	/* 0x90 */
-	0003, 0003, 0003, 0003, 0003, 0003, 0003, 0003,
-	0002, 0016, 0016, 0016, 0016, 0016, 0016, 0016,	/* 0xa0 */
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,	/* 0xb0 */
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,	/* 0xc0 */
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,	/* 0xd0 */
-	0016, 0016, 0016, 0016, 0016, 0016, 0016, 0016,
-	0013, 0013, 0013, 0013, 0013, 0013, 0013, 0013,	/* 0xe0 */
-	0013, 0013, 0013, 0013, 0013, 0013, 0013, 0013,
-	0013, 0013, 0013, 0013, 0013, 0013, 0013, 0013,	/* 0xf0 */
-	0013, 0013, 0013, 0013, 0013, 0010, 0010,    0
-};
-# endif	/* !FD */
-#endif	/* !LSI_C */
 
 static int skipdotfile = 0;
 #if	!defined (USEREGCMP) && !defined (USEREGCOMP) && !defined (USERE_COMP)
@@ -361,10 +231,12 @@ static int maxgid = 0;
 
 
 #ifndef	FD
-static VOID NEAR allocerror(VOID_A)
+VOID error(s)
+char *s;
 {
-	fputs("fatal error: memory allocation error\n", stderr);
-	fflush(stderr);
+	fputs(s, stderr);
+	fputs(": memory allocation error", stderr);
+	fputnl(stderr);
 	exit(2);
 }
 
@@ -374,7 +246,7 @@ ALLOC_T size;
 	char *tmp;
 
 	if (!size || !(tmp = (char *)malloc(size))) {
-		allocerror();
+		error("malloc()");
 #ifdef	FAKEUNINIT
 		tmp = NULL;	/* fake for -Wuninitialized */
 #endif
@@ -391,7 +263,7 @@ ALLOC_T size;
 	if (!size
 	|| !(tmp = (ptr) ? (char *)realloc(ptr, size) : (char *)malloc(size)))
 	{
-		allocerror();
+		error("realloc()");
 #ifdef	FAKEUNINIT
 		tmp = NULL;	/* fake for -Wuninitialized */
 #endif
@@ -419,7 +291,7 @@ char *s;
 
 	if (!s) return(NULL);
 	n = strlen(s);
-	if (!(tmp = (char *)malloc((ALLOC_T)n + 1))) allocerror();
+	if (!(tmp = (char *)malloc((ALLOC_T)n + 1))) error("malloc()");
 	memcpy(tmp, s, n + 1);
 	return(tmp);
 }
@@ -433,12 +305,28 @@ int n;
 
 	if (!s) return(NULL);
 	for (i = 0; i < n; i++) if (!s[i]) break;
-	if (!(tmp = (char *)malloc((ALLOC_T)i + 1))) allocerror();
+	if (!(tmp = (char *)malloc((ALLOC_T)i + 1))) error("malloc()");
 	memcpy(tmp, s, i);
 	tmp[i] = '\0';
 	return(tmp);
 }
 
+char *strchr2(s, c)
+char *s;
+int c;
+{
+	int i;
+
+	for (i = 0; s[i]; i++) {
+		if (s[i] == c) return(&(s[i]));
+		if (iskanji1(s, i)) i++;
+#ifdef	CODEEUC
+		else if (isekana(s, i)) i++;
+#endif
+	}
+	return(NULL);
+}
+
 char *strcpy2(s1, s2)
 char *s1, *s2;
 {
@@ -460,51 +348,13 @@ int n;
 	return(&(s1[i]));
 }
 
-/*
- *	ascnumeric(buf, n, 0, max): same as sprintf(buf, "%d", n)
- *	ascnumeric(buf, n, max + 1, max): same as sprintf(buf, "%0*d", max, n)
- *	ascnumeric(buf, n, max, max): same as sprintf(buf, "%*d", max, n)
- *	ascnumeric(buf, n, -1, max): same as sprintf(buf, "%-*d", max, n)
- *	ascnumeric(buf, n, x, max): like as sprintf(buf, "%*d", max, n)
- *	ascnumeric(buf, n, -x, max): like as sprintf(buf, "%-*d", max, n)
- */
-char *ascnumeric(buf, n, digit, max)
-char *buf;
-off_t n;
-int digit, max;
+# if	MSDOS
+int _dospath(path)
+char *path;
 {
-	char tmp[MAXLONGWIDTH * 2 + 1];
-	int i, j, d;
-
-	i = j = 0;
-	d = digit;
-	if (digit < 0) digit = -digit;
-	if (n < 0) tmp[i++] = '?';
-	else if (!n) tmp[i++] = '0';
-	else {
-		for (;;) {
-			tmp[i++] = '0' + n % 10;
-			if (!(n /= 10) || i >= max) break;
-			if (digit > 1 && ++j >= digit) {
-				if (i >= max - 1) break;
-				tmp[i++] = ',';
-				j = 0;
-			}
-		}
-		if (n) for (j = 0; j < i; j++) if (tmp[j] != ',') tmp[j] = '9';
-	}
-
-	if (d <= 0) j = 0;
-	else if (d > max) for (j = 0; j < max - i; j++) buf[j] = '0';
-	else for (j = 0; j < max - i; j++) buf[j] = ' ';
-	while (i--) buf[j++] = tmp[i];
-	if (d < 0) for (; j < max; j++) buf[j] = ' ';
-	buf[j] = '\0';
-
-	return(buf);
+	return((isalpha2(*path) && path[1] == ':') ? *path : 0);
 }
 
-# if	MSDOS
 DIR *Xopendir(dir)
 char *dir;
 {
@@ -569,6 +419,57 @@ DIR *dirp;
 
 	return(&d);
 }
+
+static u_int NEAR getdosmode(attr)
+u_int attr;
+{
+	u_int mode;
+
+	mode = 0;
+	if (attr & DS_IARCHIVE) mode |= S_ISVTX;
+	if (!(attr & DS_IHIDDEN)) mode |= S_IREAD;
+	if (!(attr & DS_IRDONLY)) mode |= S_IWRITE;
+	if (attr & DS_IFDIR) mode |= (S_IFDIR | S_IEXEC);
+	else if (attr & DS_IFLABEL) mode |= S_IFIFO;
+	else if (attr & DS_IFSYSTEM) mode |= S_IFSOCK;
+	else mode |= S_IFREG;
+
+	return(mode);
+}
+
+static time_t NEAR getdostime(d, t)
+u_int d, t;
+{
+	struct tm tm;
+
+	tm.tm_year = 1980 + ((d >> 9) & 0x7f);
+	tm.tm_year -= 1900;
+	tm.tm_mon = ((d >> 5) & 0x0f) - 1;
+	tm.tm_mday = (d & 0x1f);
+	tm.tm_hour = ((t >> 11) & 0x1f);
+	tm.tm_min = ((t >> 5) & 0x3f);
+	tm.tm_sec = ((t << 1) & 0x3e);
+	tm.tm_isdst = -1;
+
+	return(mktime(&tm));
+}
+
+int Xstat(path, stp)
+char *path;
+struct stat *stp;
+{
+	struct find_t find;
+
+	if (_dos_findfirst(path, SEARCHATTRS, &find) != 0) return(-1);
+	stp -> st_mode = getdosmode(find.ff_attrib);
+	stp -> st_mtime = getdostime(find.ff_fdate, find.ff_ftime);
+	stp -> st_size = find.ff_fsize;
+	stp -> st_ctime = stp -> st_atime = stp -> st_mtime;
+	stp -> st_dev = stp -> st_ino = 0;
+	stp -> st_nlink = 1;
+	stp -> st_uid = stp -> st_gid = -1;
+	return(0);
+}
 # endif	/* MSDOS */
 
 # ifdef	DJGPP
@@ -805,7 +706,7 @@ int n;
 }
 #endif	/* !PATHNOCASE */
 
-static char *NEAR getvar(ident, len)
+static char *NEAR getenvvar(ident, len)
 char *ident;
 int len;
 {
@@ -850,7 +751,13 @@ int len;
 int isidentchar(c)
 int c;
 {
-	return(c == '_' || isalpha(c));
+	return(c == '_' || isalpha2(c));
+}
+
+int isidentchar2(c)
+int c;
+{
+	return(c == '_' || isalnum2(c));
 }
 
 int isdotdir(s)
@@ -860,7 +767,6 @@ char *s;
 	return(0);
 }
 
-/*ARGSUSED*/
 char *isrootdir(s)
 char *s;
 {
@@ -931,19 +837,6 @@ int ptr, quote, len;
 #endif	/* !FAKEMETA */
 }
 
-char *long2str(s, n, size)
-char *s;
-long n;
-int size;
-{
-#ifdef	FD
-	snprintf2(s, size, "%ld", n);
-	return(s);
-#else
-	return(ascnumeric(s, (off_t)n, 0, size - 1));
-#endif
-}
-
 #ifdef	_NOORIGGLOB
 static char *NEAR cnvregexp(s, len)
 char *s;
@@ -1213,23 +1106,20 @@ int len;
 #endif
 				else paren[plen++] = s[i];
 			}
-			else {
-#ifdef	CODEEUC
-				if (isekana(s, i)) {
-					paren[plen++] = s[i++];
-					paren[plen++] = s[i];
-				}
-				else
-#endif
-				if (iskanji1(s, i)) {
-					paren[plen++] = s[i++];
-					paren[plen++] = s[i];
-				}
-#ifndef	PATHNOCASE
-				else if (!pathignorecase) paren[plen++] = s[i];
-#endif
-				else paren[plen++] = toupper2(s[i]);
+			else if (iskanji1(s, i)) {
+				paren[plen++] = s[i++];
+				paren[plen++] = s[i];
 			}
+#ifdef	CODEEUC
+			else if (isekana(s, i)) {
+				paren[plen++] = s[i++];
+				paren[plen++] = s[i];
+			}
+#endif
+#ifndef	PATHNOCASE
+			else if (!pathignorecase) paren[plen++] = s[i];
+#endif
+			else paren[plen++] = toupper2(s[i]);
 		}
 		else if (!quote && !metachar) switch (s[i]) {
 			case '?':
@@ -1250,17 +1140,16 @@ int len;
 		if (!cp) {
 			cp = malloc2(2 + 1);
 			j = 0;
-#ifdef	CODEEUC
-			if (isekana(s, i)) {
-				cp[j++] = s[i++];
-				cp[j++] = s[i];
-			}
-			else
-#endif
 			if (iskanji1(s, i)) {
 				cp[j++] = s[i++];
 				cp[j++] = s[i];
 			}
+#ifdef	CODEEUC
+			else if (isekana(s, i)) {
+				cp[j++] = s[i++];
+				cp[j++] = s[i];
+			}
+#endif
 #ifndef	PATHNOCASE
 			else if (!pathignorecase) cp[j++] = s[i];
 #endif
@@ -1293,11 +1182,10 @@ char *s;
 
 	for (n1 = n2 = 0; re[n1] && s[n2]; n1++, n2++) {
 		c1 = (u_char)(s[i = n2]);
-#ifdef	CODEEUC
-		if (isekana(s, n2)) c1 = (c1 << 8) + (u_char)(s[++n2]);
-		else
-#endif
 		if (iskanji1(s, n2)) c1 = (c1 << 8) + (u_char)(s[++n2]);
+#ifdef	CODEEUC
+		else if (isekana(s, n2)) c1 = (c1 << 8) + (u_char)(s[++n2]);
+#endif
 #ifndef	PATHNOCASE
 		else if (!pathignorecase) /*EMPTY*/;
 #endif
@@ -1331,13 +1219,12 @@ char *s;
 				i++;
 			}
 			c2 = (u_char)(re[n1][i]);
-#ifdef	CODEEUC
-			if (isekana(re[n1], i))
-				c2 = (c2 << 8) + (u_char)(re[n1][++i]);
-			else
-#endif
 			if (iskanji1(re[n1], i))
 				c2 = (c2 << 8) + (u_char)(re[n1][++i]);
+#ifdef	CODEEUC
+			else if (isekana(re[n1], i))
+				c2 = (c2 << 8) + (u_char)(re[n1][++i]);
+#endif
 			if (beg >= 0) {
 				if (beg && c1 >= beg && c1 <= c2) break;
 				beg = -1;
@@ -1808,21 +1695,23 @@ int searchhash(hpp, com, search)
 hashlist **hpp;
 char *com, *search;
 {
-	char *cp, *tmp, *next, *path;
-	int len, dlen, cost, size, ret, recalc;
 #if	MSDOS
 	char *ext;
 #endif
+	char *cp, *tmp, *next, *path;
+	int len, dlen, cost, size, ret, recalc;
 #ifndef	_NOUSEHASH
-	static char *searchcwd = NULL;
-	char buf[MAXPATHLEN];
+	hashlist *hp;
 	int n, duperrno;
 
-	if (!com && !search) {
+	if (!hpp || (!com && !search)) {
 		duperrno = errno;
-		freehash(hpp);
-		if (searchcwd) free(searchcwd);
-		searchcwd = NULL;
+		if (!com && !search) freehash(hpp);
+		if (!hpp && hashtable) for (n = 0; n < MAXHASH; n++) {
+			for (hp = hashtable[n]; hp; hp = hp -> next)
+				if (hp -> type & CM_RECALC)
+					hp -> type |= CM_REHASH;
+		}
 		errno = duperrno;
 		return(CM_NOTFOUND);
 	}
@@ -1849,11 +1738,9 @@ char *com, *search;
 	if (!hashtable) inithash();
 	n = calchash(com);
 	if ((*hpp = findhash(com, n))) {
-		path = (*hpp) -> path;
-		if (((*hpp) -> type & CM_RECALC)
-		&& (!searchcwd || !Xgetwd(buf) || strcmp(searchcwd, buf)))
-			/*EMPTY*/;
-		else if (isexecute(path, 0, 1) >= 0) return((*hpp) -> type);
+		if (!((*hpp) -> type & CM_REHASH)
+		&& isexecute((*hpp) -> path, 0, 1) >= 0)
+			return((*hpp) -> type);
 		rmhash(com, n);
 	}
 #endif
@@ -1876,8 +1763,6 @@ char *com, *search;
 		*hpp = newhash(com, path, 0, hashtable[n]);
 		hashtable[n] = *hpp;
 		(*hpp) -> type = (ret |= (CM_HASH | CM_RECALC));
-		if (searchcwd) free(searchcwd);
-		searchcwd = strdup2(Xgetwd(buf));
 # endif
 		return(ret);
 	}
@@ -1927,8 +1812,6 @@ char *com, *search;
 			*hpp = newhash(com, path, cost, hashtable[n]);
 			hashtable[n] = *hpp;
 			(*hpp) -> type = (ret |= (CM_HASH | recalc));
-			if (searchcwd) free(searchcwd);
-			searchcwd = strdup2(Xgetwd(buf));
 #endif
 			return(ret);
 		}
@@ -1938,7 +1821,7 @@ char *com, *search;
 	return(CM_NOTFOUND);
 }
 
-char *searchpath(path, search)
+char *searchexecpath(path, search)
 char *path, *search;
 {
 	hashlist *hp;
@@ -2183,6 +2066,11 @@ int stripm, quoted;
 			if (s1) s1[j] = PMETA;
 			j++;
 		}
+		else if (iskanji1(s2, i)) {
+			if (s1) s1[j] = s2[i];
+			j++;
+			i++;
+		}
 #ifdef	CODEEUC
 		else if (isekana(s2, i)) {
 			if (s1) s1[j] = s2[i];
@@ -2190,11 +2078,6 @@ int stripm, quoted;
 			i++;
 		}
 #endif
-		else if (iskanji1(s2, i)) {
-			if (s1) s1[j] = s2[i];
-			j++;
-			i++;
-		}
 		if (s1) s1[j] = s2[i];
 	}
 	return(j);
@@ -2243,6 +2126,24 @@ char **var;
 	errno = duperrno;
 }
 
+char **duplvar(var, margin)
+char **var;
+int margin;
+{
+	char **dupl;
+	int i, n;
+
+	if (margin < 0) {
+		if (!var) return(NULL);
+		margin = 0;
+	}
+	n = countvar(var);
+	dupl = (char **)malloc2((n + margin + 1) * sizeof(char *));
+	for (i = 0; i < n; i++) dupl[i] = strdup2(var[i]);
+	dupl[i] = NULL;
+	return(dupl);
+}
+
 int parsechar(s, len, spc, qflag, qp, pqp)
 char *s;
 int len, spc, qflag, *qp, *pqp;
@@ -2255,10 +2156,10 @@ int len, spc, qflag, *qp, *pqp;
 		}
 		return(PC_CLQUOTE);
 	}
+	if (iskanji1(s, 0)) return(PC_WORD);
 #ifdef	CODEEUC
 	if (isekana(s, 0)) return(PC_WORD);
 #endif
-	if (iskanji1(s, 0)) return(PC_WORD);
 	if (*qp == '`') return(PC_BQUOTE);
 	if (*qp == '\'') return(PC_SQUOTE);
 	if (spc && *s == spc) return(*s);
@@ -2292,8 +2193,7 @@ int *eolp, *ptrp, qed;
 
 	if (isidentchar((*bufp)[*ptrp])) {
 		while ((*bufp)[++(*ptrp)])
-			if (!isidentchar((*bufp)[*ptrp])
-			&& !isdigit((*bufp)[*ptrp])) break;
+			if (!isidentchar2((*bufp)[*ptrp])) break;
 		if (eolp) *eolp = *ptrp;
 		return(mode);
 	}
@@ -2329,8 +2229,7 @@ int *eolp, *ptrp, qed;
 #endif
 	if (isidentchar((*bufp)[*ptrp])) {
 		while ((*bufp)[++(*ptrp)])
-			if (!isidentchar((*bufp)[*ptrp])
-			&& !isdigit((*bufp)[*ptrp])) break;
+			if (!isidentchar2((*bufp)[*ptrp])) break;
 	}
 	else (*ptrp)++;
 
@@ -2595,21 +2494,27 @@ int plen, *modep;
 			break;
 		case '#':
 			if (!argvar) break;
-			cp = long2str(tmp, countvar(argvar + 1), sizeof(tmp));
+			snprintf2(tmp, sizeof(tmp), "%d",
+				countvar(argvar + 1));
+			cp = tmp;
 			break;
 		case '?':
-			i = (getretvalfunc) ? (*getretvalfunc)() : 0;
-			cp = long2str(tmp, i, sizeof(tmp));
+			snprintf2(tmp, sizeof(tmp), "%d",
+				(getretvalfunc) ? (*getretvalfunc)() : 0);
+			cp = tmp;
 			break;
 		case '$':
 			if (!getpidfunc) pid = getpid();
 			else pid = (*getpidfunc)();
-			cp = long2str(tmp, pid, sizeof(tmp));
+			snprintf2(tmp, sizeof(tmp), "%id", pid);
+			cp = tmp;
 			break;
 		case '!':
 			if (getlastpidfunc
-			&& (pid = (*getlastpidfunc)()) >= 0)
-				cp = long2str(tmp, pid, sizeof(tmp));
+			&& (pid = (*getlastpidfunc)()) >= 0) {
+				snprintf2(tmp, sizeof(tmp), "%id", pid);
+				cp = tmp;
+			}
 			break;
 		case '-':
 			if (getflagfunc) cp = (*getflagfunc)();
@@ -2669,14 +2574,11 @@ int s, len, vlen, mode, quoted;
 		demacroarg(cpp);
 #endif
 		for (i = 0; i < len; i++) fputc(arg[i], stderr);
-		fputs(": ", stderr);
-		if (vlen <= 0)
-			fputs("parameter null or not set", stderr);
-		else for (i = 0; (*cpp)[i]; i++) fputc((*cpp)[i], stderr);
+		fprintf2(stderr, ": %k",
+			(vlen > 0) ? *cpp : "parameter null or not set");
+		fputnl(stderr);
 		free(*cpp);
 		*cpp = NULL;
-		fputc('\n', stderr);
-		fflush(stderr);
 		if (exitfunc) (*exitfunc)();
 		return(-2);
 	}
@@ -2753,8 +2655,8 @@ int quoted;
 
 	c = '\0';
 	if ((cp = new)) /*EMPTY*/;
-	else if (isidentchar(*top)) cp = getvar(top, len);
-	else if (isdigit(*top)) {
+	else if (isidentchar(*top)) cp = getenvvar(top, len);
+	else if (isdigit2(*top)) {
 		if (len > 1) return(-1);
 		if (!argvar) {
 			(*bufp)[ptr++] = '$';
@@ -2786,8 +2688,11 @@ int quoted;
 	if (cp && (nul == 0x80) && !*cp) cp = NULL;
 
 #ifndef	MINIMUMSHELL
-	if (mode == 'n')
-		cp = long2str(tmp, (cp) ? strlen(cp) : 0, sizeof(tmp));
+	if (mode == 'n') {
+		snprintf2(tmp, sizeof(tmp), "%d",
+			(cp) ? strlen(cp) : 0);
+		cp = tmp;
+	}
 	else if (nul == -1) {
 		new = removeword(cp, top + s, vlen, mode);
 		if (new) cp = new;
@@ -2917,7 +2822,7 @@ char *name;
 	gidlist = b_realloc(gidlist, maxgid, gidtable);
 	gidlist[maxgid].gid = grp -> gr_gid;
 	gidlist[maxgid].name = strdup2(grp -> gr_name);
-	gidlist[maxgid].gr_mem = grp -> gr_mem;
+	gidlist[maxgid].gr_mem = duplvar(grp -> gr_mem, -1);
 	gidlist[maxgid].ismem = 0;
 	return(&(gidlist[maxgid++]));
 }
@@ -2932,13 +2837,15 @@ gid_t gid;
 	if (!(gp = findgid(gid, NULL))) return(0);
 	if (!(gp -> ismem)) {
 		gp -> ismem++;
-		if ((up = finduid(geteuid(), NULL)))
+		if (gp -> gr_mem && (up = finduid(geteuid(), NULL)))
 		for (i = 0; gp -> gr_mem[i]; i++) {
 			if (!strpathcmp(up -> name, gp -> gr_mem[i])) {
 				gp -> ismem++;
 				break;
 			}
 		}
+		freevar(gp -> gr_mem);
+		gp -> gr_mem = NULL;
 	}
 
 	return(gp -> ismem - 1);
@@ -2957,7 +2864,10 @@ VOID freeidlist(VOID_A)
 		free(uidlist);
 	}
 	if (gidlist) {
-		for (i = 0; i < maxgid; i++) free(gidlist[i].name);
+		for (i = 0; i < maxgid; i++) {
+			free(gidlist[i].name);
+			freevar(gidlist[i].gr_mem);
+		}
 		free(gidlist);
 	}
 	uidlist = NULL;
@@ -3021,6 +2931,7 @@ char *gethomedir(VOID_A)
 	return(NULL);
 }
 
+#ifndef	MINIMUMSHELL
 int evalhome(bufp, ptr, argp)
 char **bufp;
 int ptr;
@@ -3033,26 +2944,26 @@ char **argp;
 
 	len = ((cp = strdelim(top, 0))) ? (cp - top) : strlen(top);
 	if (!len) cp = gethomedir();
-#ifdef	FD
+# ifdef	FD
 	else if (len == sizeof("FD") - 1 && !strnpathcmp(top, "FD", len))
 		cp = progpath;
-#endif
+# endif
 	else {
-#ifdef	NOUID
+# ifdef	NOUID
 		cp = NULL;
-#else	/* !NOUID */
-# ifdef	FD
+# else	/* !NOUID */
+#  ifdef	FD
 		uidtable *up;
 
 		cp = strndup2(top, len);
 		up = finduid(0, cp);
 		free(cp);
 		cp = (up) ? up -> home : NULL;
-# else	/* !FD */
+#  else	/* !FD */
 		struct passwd *pwd;
 
 		cp = strndup2(top, len);
-#  ifdef	DEBUG
+#   ifdef	DEBUG
 		_mtrace_file = "getpwnam(start)";
 		pwd = getpwnam(cp);
 		if (_mtrace_file) _mtrace_file = NULL;
@@ -3060,13 +2971,13 @@ char **argp;
 			_mtrace_file = "getpwnam(end)";
 			malloc(0);	/* dummy alloc */
 		}
-#  else
+#   else
 		pwd = getpwnam(cp);
-#  endif
+#   endif
 		free(cp);
 		cp = (pwd) ? pwd -> pw_dir : NULL;
-# endif	/* !FD */
-#endif	/* !NOUID */
+#  endif	/* !FD */
+# endif	/* !NOUID */
 	}
 
 	if (!cp) {
@@ -3083,6 +2994,7 @@ char **argp;
 	*argp += len;
 	return(ptr);
 }
+#endif	/* !MINIMUMSHELL */
 
 char *evalarg(arg, stripq, backq, qed)
 char *arg;
@@ -3157,7 +3069,11 @@ int stripq, backq, qed;
 			tmpq = (q) ? q : qed;
 			if (!cp[1]) buf[i++] = *cp;
 #ifdef	FAKEMETA
+# ifdef	MINIMUMSHELL
+			else if (cp[1] == '$') cp++;
+# else
 			else if (cp[1] == '$' || cp[1] == '~') cp++;
+# endif
 #endif
 			else if ((i = evalvar(&buf, i, &cp, tmpq)) < 0) {
 				if (bbuf) free(bbuf);
@@ -3180,8 +3096,10 @@ int stripq, backq, qed;
 			else if (!stripq) buf[i++] = *cp;
 		}
 		else if (pc != PC_NORMAL) /*EMPTY*/;
+#ifndef	MINIMUMSHELL
 		else if (*cp == '~' && (!prev || prev == ':' || prev == '='))
 			i = evalhome(&buf, i, &cp);
+#endif
 		else buf[i++] = *cp;
 	}
 #ifndef	BASHSTYLE
@@ -3209,9 +3127,9 @@ char ***argvp, *ifs;
 				'\0', 0, &quote, NULL);
 			if (pc == PC_WORD || pc == PC_META) i++;
 			else if (pc != PC_NORMAL) /*EMPTY*/;
-			else if (strchr(ifs, (*argvp)[n][i])) {
+			else if (strchr2(ifs, (*argvp)[n][i])) {
 				for (j = i + 1; (*argvp)[n][j]; j++)
-					if (!strchr(ifs, (*argvp)[n][j]))
+					if (!strchr2(ifs, (*argvp)[n][j]))
 						break;
 				if (!i) {
 					for (i = 0; (*argvp)[n][i + j]; i++)
@@ -3237,7 +3155,7 @@ char ***argvp, *ifs;
 			free((*argvp)[n--]);
 			memmove((char *)(&((*argvp)[n + 1])),
 				(char *)(&((*argvp)[n + 2])),
-				(--argc - n) * sizeof(char *));
+				(argc-- - n) * sizeof(char *));
 		}
 	}
 	return(argc);
