@@ -5,9 +5,7 @@
  */
 
 #include "fd.h"
-#include "term.h"
 #include "func.h"
-#include "kctype.h"
 #include "kanji.h"
 
 #if	MSDOS && !defined (_NOUSELFN)
@@ -52,7 +50,7 @@ static int NEAR setarg __P_((char **, int, ALLOC_T *, char *, char *, int));
 static int NEAR flag2str __P_((char *, int, int));
 static int NEAR skipquote __P_((char *, int *));
 #if	!defined (MACROMETA) || !defined (_NOEXTRAMACRO)
-static char *NEAR restorearg __P_((char *));
+static char *NEAR _restorearg __P_((char *));
 static char *NEAR _demacroarg __P_((char *));
 #endif
 #ifdef	_NOEXTRAMACRO
@@ -132,9 +130,8 @@ int code;
 	}
 # endif
 	(*bufp)[eol] = '\0';
-	len = (eol - ptr) * 3 + 3;
-	cp = malloc2(len + 1);
-	len = kanjiconv(cp, &((*bufp)[ptr]), len, DEFCODE, code, L_FNAME);
+	cp = newkanjiconv(&((*bufp)[ptr]), DEFCODE, code, L_FNAME);
+	len = strlen(cp);
 
 	*bufp = c_realloc(*bufp, ptr + len + 1, sizep);
 	memcpy(&((*bufp)[ptr]), cp, len);
@@ -214,7 +211,8 @@ int flags;
 
 	if (arg != path) arg = convput(conv, arg, 1, 0, NULL, NULL);
 #ifndef	_NOKANJIFCONV
-	else arg = kanjiconv2(conv, arg, MAXPATHLEN - 1, DEFCODE, fnamekcode);
+	else arg = kanjiconv2(conv, arg, MAXPATHLEN - 1,
+		DEFCODE, fnamekcode, L_FNAME);
 #endif
 	optr = ptr;
 	ptr = checksc(*bufp, ptr, arg);
@@ -271,12 +269,13 @@ int *ptrp;
 }
 
 #if	!defined (MACROMETA) || !defined (_NOEXTRAMACRO)
-static char *NEAR restorearg(s)
+static char *NEAR _restorearg(s)
 char *s;
 {
 	char *buf;
 	int i, j, m, flags;
 
+	if (!s) return(NULL);
 	i = strlen(s);
 # ifdef	MACROMETA
 	buf = malloc2(i * 3 + 1);	/* MACROMETA+flags -> %XSTA (x2.5) */
@@ -311,10 +310,10 @@ char *s;
 static char *NEAR _demacroarg(s)
 char *s;
 {
-	char *buf;
+	char *cp;
 
-	if ((buf = restorearg(s)) != s) free(s);
-	return(buf);
+	if ((cp = _restorearg(s)) != s && s) free(s);
+	return(cp);
 }
 #endif	/* !MACROMETA || !_NOEXTRAMACRO */
 
@@ -643,6 +642,7 @@ int ignorelist;
 				else if (c == 'H') tmpcode = HEX;
 				else if (c == 'C') tmpcode = CAP;
 				else if (c == 'U') tmpcode = UTF8;
+				else if (c == 'M') tmpcode = M_UTF8;
 #  ifndef	_NOKANJIFCONV
 				else if (c == 'A') tmpcode = -1;
 #  endif
@@ -1091,42 +1091,24 @@ char **argp;
 }
 #endif	/* !_NOEXTRAMACRO */
 
-#if	!defined (MACROMETA) || defined (_NOEXTRAMACRO)
-int argfputs(s, fp)
+char *restorearg(s)
 char *s;
-FILE *fp;
 {
-	return(kanjifputs(s, fp));
+#if	defined (MACROMETA) && !defined (_NOEXTRAMACRO)
+	return(_restorearg(s));
+#else
+	return(s);
+#endif
 }
 
 /*ARGSUSED*/
 VOID demacroarg(argp)
 char **argp;
 {
-	return;
-}
-
-#else	/* MACROMETA && !_NOEXTRAMACRO */
-
-int argfputs(s, fp)
-char *s;
-FILE *fp;
-{
-	char *buf;
-	int n;
-
-	buf = restorearg(s);
-	n = kanjifputs(buf, fp);
-	if (buf != s) free(buf);
-	return(n);
-}
-
-VOID demacroarg(argp)
-char **argp;
-{
+#if	defined (MACROMETA) && !defined (_NOEXTRAMACRO)
 	*argp = _demacroarg(*argp);
+#endif
 }
-#endif	/* MACROMETA && !_NOEXTRAMACRO */
 
 char *inputshellstr(prompt, ptr, def)
 char *prompt;
@@ -1271,10 +1253,7 @@ int ignorelist;
 	if (!wastty) stdiomode();
 	free(command);
 	if (!cp) {
-		if (!wastty) {
-			fputc('\n', stdout);
-			fflush(stdout);
-		}
+		if (!wastty) fputnl(stdout);
 		return(NULL);
 	}
 	if (!*cp) {
@@ -1495,7 +1474,7 @@ char *argv[];
 		memcpy(&(line[j]), &(command[i]), len);
 		i += len + 1;
 		j += len;
-		if (!isdigit(command[i])) {
+		if (!isdigit2(command[i])) {
 			line[j++] = '$';
 			line[j++] = command[i++];
 		}
@@ -1572,12 +1551,12 @@ int uniq;
 {
 	int i, size;
 
-	size = histsize[n];
+	size = (int)histsize[n];
 	if (!history[n]) {
 		history[n] = (char **)malloc2(sizeof(char *) * (size + 1));
 		for (i = 0; i <= size; i++) history[n][i] = NULL;
 		histbufsize[n] = size;
-		histno[n] = 0;
+		histno[n] = (short)0;
 	}
 	else if (size > histbufsize[n]) {
 		history[n] = (char **)realloc2(history[n],
@@ -1589,7 +1568,7 @@ int uniq;
 
 	if (!s || !*s) return(0);
 
-	if (histno[n]++ >= MAXHISTNO) histno[n] = 0;
+	if (histno[n]++ >= MAXHISTNO) histno[n] = (short)0;
 
 	if (uniq) {
 		for (i = 0; i <= size; i++) {
@@ -1614,9 +1593,9 @@ int n;
 	char *tmp;
 	int i, size;
 
-	size = histsize[n];
+	size = (int)histsize[n];
 	if (!history[n] || size > histbufsize[n] || size <= 0) return(NULL);
-	if (--histno[n] < 0) histno[n] = MAXHISTNO;
+	if (--histno[n] < (short)0) histno[n] = MAXHISTNO;
 	tmp = history[n][0];
 	for (i = 0; i < size; i++) history[n][i] = history[n][i + 1];
 	history[n][size--] = NULL;
@@ -1633,14 +1612,14 @@ char *file;
 
 	if (!(fp = Xfopen(file, "r"))) return(-1);
 
-	size = histsize[n];
+	size = (int)histsize[n];
 	history[n] = (char **)malloc2(sizeof(char *) * (size + 1));
 	histbufsize[n] = size;
-	histno[n] = 0;
+	histno[n] = (short)0;
 
 	i = -1;
 	while ((line = fgets2(fp, 1))) {
-		if (histno[n]++ >= MAXHISTNO) histno[n] = 0;
+		if (histno[n]++ >= MAXHISTNO) histno[n] = (short)0;
 		if (i < size) i++;
 		else free(history[n][i]);
 		for (j = i; j > 0; j--) history[n][j] = history[n][j - 1];
@@ -1663,7 +1642,7 @@ char *file;
 	if (!history[n] || !history[n][0]) return(-1);
 	if (!(fp = Xfopen(file, "w"))) return(-1);
 
-	size = (savehist > histsize[n]) ? histsize[n] : savehist;
+	size = (savehist > (int)histsize[n]) ? (int)histsize[n] : savehist;
 	for (i = size - 1; i >= 0; i--) if (history[n][i] && *history[n][i]) {
 		for (cp = history[n][i]; (eol = strchr(cp, '\n')); cp = eol) {
 			Xfwrite(cp, sizeof(char), eol++ - cp, fp);
@@ -1682,22 +1661,20 @@ char *str;
 int *ptrp;
 {
 	char *cp;
-	long no;
 	int i, n, ptr, size;
 
 	ptr = (ptrp) ? *ptrp : 0;
-	size = histsize[0];
+	size = (int)histsize[0];
 	if (str[ptr] == '!') n = 0;
 	else if (str[ptr] == '=' || strchr(IFS_SET, str[ptr])) n = ptr = -1;
-	else if ((cp = evalnumeric(&(str[ptr]), &no, -1))) {
-		if (!no) n = -1;
-		else if (no < 0) {
-			if (-1L - no > (long)MAXHISTNO) n = -1;
-			else n = -1L - no;
+	else if ((cp = sscanf2(&(str[ptr]), "%-d", &n))) {
+		if (!n) n = -1;
+		else if (n < 0) {
+			if ((n = -1 - n) > (int)MAXHISTNO) n = -1;
 		}
-		else if (no - 1L > (long)MAXHISTNO) n = -1;
-		else if (no <= (long)(histno[0])) n = histno[0] - no;
-		else n = MAXHISTNO - (no - histno[0]) + 1;
+		else if (n - 1 > (int)MAXHISTNO) n = -1;
+		else if (n <= (int)(histno[0])) n = (int)histno[0] - n;
+		else n = MAXHISTNO - (n - (int)histno[0]) + 1;
 		ptr = (cp - str) - 1;
 	}
 	else {
@@ -1746,8 +1723,8 @@ char *command;
 					continue;
 				}
 				command[++i] = '\0';
-				fputs(&(command[len]), stderr);
-				fputs(": Event not found.\n", stderr);
+				fprintf2(stderr, "%k: Event not found.\n",
+					&(command[len]));
 				free(cp);
 				return(NULL);
 			}
@@ -1763,7 +1740,7 @@ char *command;
 	cp[j] = '\0';
 	if (hit) {
 		kanjifputs(cp, stderr);
-		fputc('\n', stderr);
+		fputnl(stderr);
 		free(command);
 		return(cp);
 	}
